\name{spacegraph}
\alias{spacegraph}
\title{Randomly compute many different matching solutions}
\description{
Randomly compute many different matching solutions
}
\usage{
spacegraph(treatment=NULL, data = NULL, drop=NULL, 
     R=list("cem"=50,"psm"=0,"mdm"=0,"matchit"=0), 
     grouping = NULL,  L1.breaks = NULL, L1.grouping=NULL, 
     fixed = NULL, minimal = 1, maximal = 15, M=100, 
     raw.profile=NULL, keep.weights=FALSE, 
     rgrouping=F,groups=NULL,psmpoly=1,mdmpoly=1, 
     other.matches=NULL, heuristic=F, progress=TRUE,) 

}
\arguments{
  \item{treatment}{character, name of the treatment variable}
  \item{data}{a data.frame}
  \item{drop}{a vector of variable names in the data frame to ignore 
    during matching.}
  \item{R}{a named list that gives the number of of possible random solutions for each matching method.  Allowed methods are \code{cem}, \code{psm}, \code{mdm}, and \code{matchit}.}
  \item{grouping}{named list, each element of which is a list of
  groupings for a single categorical variable. For more details see \code{\link{cem}}.}
  \item{L1.breaks}{list of cutpoints for the calculation of the L1 measure.}
  \item{L1.grouping}{as \code{grouping} but only needed in the calculation
   of the L1 measure not in matching.}
  \item{fixed}{vector of variable names which will not be relaxed.} 
  \item{minimal}{the minimal number of intervals acceptable after
  relaxation. Should be a named list of positive integers or if a number, this is
  applied to all variables.}
  \item{maximal}{the maximal number of intervals acceptable after
  relaxation. Should be a named list of positive integers or if a number, this is
  applied to all variables.}
  \item{M}{number of possible random coarsening for the L1 measure}
  \item{raw.profile}{an object of class \code{L1profile}. If passed, the \code{L1.breaks} are
  ignored and set to median cutpoints of L1 profile.}
  \item{keep.weights}{if \code{TRUE}, for each matching solutions the CEM-weights are stored.}
 
  \item{rgrouping}{Boolean, specifies whether levels of categorical variables should be randomly grouped together by CEM.}
  \item{groups}{named list, each element of which is a list of
    allowable groupings for a single categorical variable.}
  \item{psmpoly}{numeric, specifying the order of polynomials to include in the propensity score models.  At the moment, only \code{psmpoly=1} is available and other values will throw warnings.}
  \item{mdmpoly}{numeric, specifying the order of polynomials to include in Mahalanobis matching.  At the moment, only \code{mdmpoly=1} is available and other values will throw warnings.}
  \item{other.matches}{This argument allows plotting of user-specified matching solutions.  The solutions must be supplied in a specific format: as a list of data frames, where each data frame contains the observation IDs, observations weights, and the method.  See the example.}
  \item{heuristic}{Boolean, if TRUE spacegraph uses a heuristic method to randomly select covariates for MDM and PSM rather than sampling from all possible combinations of covariates.  The heuristic tends to select most of the main effects and a small number of interactions.  This method is automatically applied with large numbers of covariates.}
  \item{progress}{show progress bars.}


}

\details{
Spacegraph is a tool to help the user to the search for optimal matching solutions by generating many matching solutions from a variety of matching algorithms (currently CEM, Mahalanobis distance matching, and propensity score matching are supported).  The resulting object can be plotted with \code{plot()} to show where each solution falls along the bais-variance tradeoff.

} 
\value{
  \item{val}{an object of class \code{imbalance.space} that can be used directly with \code{plot()} to produce a spacegraph.}
}

\references{}
\author{Richard Nielsen, Stephano Iacus, Gary King, and Guiseppe Porro}
\seealso{ \code{\link{cemspace}} }
\examples{
data(LL)

sp <- spacegraph("treated", LL, drop="re78", M=5,
                  R=list(cem=5,psm=5, mdm=5))

plot(sp)
## ABOUT THE PLOTTING TOOL:
## The circled solution is the current selection.
## Solutions that are strictly better are also circled.
## The gui provides the exact call to re-run the selected matching solution.
## The call can also be edited, re-run by clicking the button, and 
## automatically added to the existing spacegraph.
## CEM solutions can also be adjusted variable by variable and re-run.


## Some plotting parameters can be changed
plot(sp, main="Comparison of Matching Methods",
  ylab="L1", xlim=c(300,50), ylim=c(0,.7))

}
\keyword{datagen}

